require('skyscrapers')

local skyscrapers = getSkyscraperDefinitions()
local maxSkyscraperSizeX = getBiggestSkyscraperX()
local maxSkyscraperSizeY = getBiggestSkyscraperY()
local maxSkyscraperSizeXY = math.max(maxSkyscraperSizeX,maxSkyscraperSizeY)
local spacingX = getSpacingX()
local spacingY = getSpacingY()
local premadeSkyscrapers = getPremadeSkyscraperDefinitions()

data:extend({
    
    {
        type = "item-group",
        name = "skyscrapers",
        order = "b-m",
        inventory_order = "d-m",
        icon = "__Skyscrapers__/graphics/icons/skyscraper_big.png",
        icon_size = 96,
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-buildings-mk1",
        group = "skyscrapers",
        order = "a-a"
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-buildings-mk2",
        group = "skyscrapers",
        order = "b-a"
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-buildings-mk3",
        group = "skyscrapers",
        order = "c-a"
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-buildings-unique",
        group = "skyscrapers",
        order = "d-a"
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-buildings-premade",
        group = "skyscrapers",
        order = "e-a"
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-floors",
        group = "skyscrapers",
        order = "f-a"
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-elevators",
        group = "skyscrapers",
        order = "g-a"
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-cables",
        group = "skyscrapers",
        order = "h-a"
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-beacons",
        group = "skyscrapers",
        order = "h-a"
    },
    {
        type = "item-subgroup",
        name = "skyscrapers-belts",
        group = "skyscrapers",
        order = "i-a"
    },
    })

local customOutOfMap = table.deepcopy(data.raw["tile"]["grass-1"])
customOutOfMap.collision_mask = data.raw["tile"]["out-of-map"].collision_mask
customOutOfMap.variants = data.raw["tile"]["out-of-map"].variants
customOutOfMap.name = "skyscraper-boundary"
data:extend({customOutOfMap})

--todo consider omitting them if 5dim is present (or done generically if a belt of sufficient length exists)
local longUndergrounds1 = table.deepcopy(data.raw["underground-belt"]["underground-belt"])
longUndergrounds1.name = "long-underground-belt"
longUndergrounds1.minable.result = "long-underground-belt"
longUndergrounds1.max_distance = 50
data:extend({longUndergrounds1})
local longUndergroundsItem1 = table.deepcopy(data.raw["item"]["underground-belt"])
longUndergroundsItem1.name = "long-underground-belt"
longUndergroundsItem1.place_result = "long-underground-belt"
longUndergroundsItem1.subgroup="skyscrapers-belts"
data:extend({longUndergroundsItem1})
local longUndergroundsRecipe1 = table.deepcopy(data.raw["recipe"]["underground-belt"])
longUndergroundsRecipe1.name = "long-underground-belt"
longUndergroundsRecipe1.ingredients =
{
  {"iron-plate", 10},
  {"transport-belt", 50}
}
longUndergroundsRecipe1.result = "long-underground-belt"
data:extend({longUndergroundsRecipe1})
table.insert(data.raw["technology"]["logistics"].effects, {
    type = "unlock-recipe",
    recipe = "long-underground-belt"
})

local longUndergrounds2 = table.deepcopy(data.raw["underground-belt"]["fast-underground-belt"])
longUndergrounds2.name = "long-fast-underground-belt"
longUndergrounds2.minable.result = "long-fast-underground-belt"
longUndergrounds2.max_distance = 50
data:extend({longUndergrounds2})
local longUndergroundsItem2 = table.deepcopy(data.raw["item"]["fast-underground-belt"])
longUndergroundsItem2.name = "long-fast-underground-belt"
longUndergroundsItem2.place_result = "long-fast-underground-belt"
longUndergroundsItem2.subgroup="skyscrapers-belts"
data:extend({longUndergroundsItem2})
local longUndergroundsRecipe2 = table.deepcopy(data.raw["recipe"]["fast-underground-belt"])
longUndergroundsRecipe2.name = "long-fast-underground-belt"
longUndergroundsRecipe2.ingredients =
{
  {"iron-plate", 10},
  {"fast-transport-belt", 50}
}
longUndergroundsRecipe2.result = "long-fast-underground-belt"
data:extend({longUndergroundsRecipe2})

local longUndergrounds3 = table.deepcopy(data.raw["underground-belt"]["express-underground-belt"])
longUndergrounds3.name = "long-express-underground-belt"
longUndergrounds3.minable.result = "long-express-underground-belt"
longUndergrounds3.max_distance = 50
data:extend({longUndergrounds3})
local longUndergroundsItem3 = table.deepcopy(data.raw["item"]["express-underground-belt"])
longUndergroundsItem3.name = "long-express-underground-belt"
longUndergroundsItem3.place_result = "long-express-underground-belt"
longUndergroundsItem3.subgroup="skyscrapers-belts"
data:extend({longUndergroundsItem3})
local longUndergroundsRecipe3 = table.deepcopy(data.raw["recipe"]["express-underground-belt"])
longUndergroundsRecipe3.name = "long-express-underground-belt"
longUndergroundsRecipe3.ingredients =
{
  {"iron-plate", 10},
  {"express-transport-belt", 50}
}
longUndergroundsRecipe3.result = "long-express-underground-belt"
data:extend({longUndergroundsRecipe3})
table.insert(data.raw["technology"]["logistics-3"].effects, {
    type = "unlock-recipe",
    recipe = "long-express-underground-belt"
})

local longPipeToGround = table.deepcopy(data.raw["pipe-to-ground"]["pipe-to-ground"])
longPipeToGround.name = "long-pipe-to-ground"
longPipeToGround.minable.result = "long-pipe-to-ground"
longPipeToGround.fluid_box.pipe_connections[2].max_underground_distance = 50
data:extend({longPipeToGround})
local longPipeToGroundItem = table.deepcopy(data.raw["item"]["pipe-to-ground"])
longPipeToGroundItem.name = "long-pipe-to-ground"
longPipeToGroundItem.place_result = "long-pipe-to-ground"
longPipeToGroundItem.subgroup="skyscrapers-belts"
data:extend({longPipeToGroundItem})
local longPipeToGroundRecipe = table.deepcopy(data.raw["recipe"]["pipe-to-ground"])
longPipeToGroundRecipe.name = "long-pipe-to-ground"
longPipeToGroundRecipe.ingredients =
{
  {"iron-plate", 5},
  {"pipe", 50}
}
longPipeToGroundRecipe.result = "long-pipe-to-ground"
longPipeToGroundRecipe.enabled=false
data:extend({longPipeToGroundRecipe})

local substation = table.deepcopy(data.raw["electric-pole"]["substation"])
substation.name = "skyscrapers-substation"
substation.order = "nowhere"
substation.minable.result = nil
substation.maximum_wire_distance = 1
substation.supply_area_distance = math.ceil(1+maxSkyscraperSizeXY/2.0)
substation.connection_points = {{ shadow = { copper = {0, 999}, green = {0, 999}, red = {0, 999}}, wire ={ copper = {0, 999}, green = {0, 999}, red = {0, 999}}}}
substation.pictures.direction_count = 1
substation.pictures.filename = "__Skyscrapers__/graphics/transparent.png"
substation.pictures.width = 32
substation.pictures.height = 32
substation.pictures.shift = {0.0, 0.0}
substation.collision_box = {{0.0, 0.0}, {0.0, 0.0}}
substation.selection_box = {{0, 0}, {0, 0}}
substation.collision_mask = {}
table.insert(substation.flags,"not-on-map")
table.insert(substation.flags,"not-blueprintable")
data:extend({substation})

local pole = table.deepcopy(data.raw["electric-pole"]["medium-electric-pole"])
pole.name = "skyscrapers-pole"
pole.order = "nowhere"
pole.minable.result = nil
pole.maximum_wire_distance = math.ceil(spacingY / 2.0)
pole.supply_area_distance = 1
pole.connection_points = {{ shadow = { copper = {0, 999}, green = {0, 999}, red = {0, 999}}, wire ={ copper = {0, 999}, green = {0, 999}, red = {0, 999}}}}
pole.pictures.direction_count = 1
pole.pictures.filename = "__Skyscrapers__/graphics/transparent.png"
pole.pictures.width = 32
pole.pictures.height = 32
pole.pictures.shift = {0.0, 0.0}
pole.collision_box = {{0.0, 0.0}, {0.0, 0.0}}
pole.selection_box = {{0, 0}, {0, 0}}
pole.collision_mask = {}
table.insert(pole.flags,"not-on-map")
table.insert(pole.flags,"not-blueprintable")
data:extend({pole})

local beaconTemplate = table.deepcopy(data.raw["beacon"]["beacon"])
beaconTemplate.name = "skyscrapers-beacon-template"
beaconTemplate.order = "nowhere"
beaconTemplate.minable.result = nil
beaconTemplate.supply_area_distance = 0
beaconTemplate.localised_name = {"entity-name.skyscrapers-beacon-template"}
beaconTemplate.localised_description = {"entity-description.skyscrapers-beacon-template"}
table.insert(beaconTemplate.flags,"not-blueprintable")


local beaconInstance = table.deepcopy(data.raw["beacon"]["beacon"])
beaconInstance.name = "skyscrapers-beacon-instance"
beaconInstance.order = "nowhere"
beaconInstance.minable.result = "beacon"
beaconInstance.supply_area_distance = math.ceil(1+maxSkyscraperSizeXY/2.0)
--beaconInstance.collision_box = {{0.0, 0.0}, {0.0, 0.0}} --for some reason if the bounding box is empty, the beacon no longer affects entities
beaconInstance.selection_box = {{0, 0}, {0, 0}}
beaconInstance.collision_mask = {}
table.insert(beaconInstance.flags,"not-on-map")
table.insert(beaconInstance.flags,"not-blueprintable")
beaconInstance.base_picture.filename = "__Skyscrapers__/graphics/transparent.png"
beaconInstance.base_picture.width = 32
beaconInstance.base_picture.height = 32
beaconInstance.base_picture.shift = {0.0, 0.0}
beaconInstance.animation.filename = "__Skyscrapers__/graphics/transparent.png"
beaconInstance.animation.width = 32
beaconInstance.animation.height = 32
beaconInstance.animation.shift = {0.0, 0.0}
beaconInstance.animation.line_length = 1
beaconInstance.animation.frame_count = 1
beaconInstance.animation.animation_speed = 1
beaconInstance.animation_shadow.filename = "__Skyscrapers__/graphics/transparent.png"
beaconInstance.animation_shadow.width = 32
beaconInstance.animation_shadow.height = 32
beaconInstance.animation_shadow.shift = {0.0, 0.0}
beaconInstance.animation_shadow.line_length = 1
beaconInstance.animation_shadow.frame_count = 1
beaconInstance.animation_shadow.animation_speed = 1
beaconInstance.energy_usage = "100W" --instead all consumptions are calculated for the template beacon. however 0W results in a constantly flashing power icon so some consumption is needed

local elevator = table.deepcopy(data.raw["car"]["car"])
elevator.name = "skyscrapers-elevator"
elevator.order = "nowhere"
elevator.minable.result = nil
elevator.collision_box = {{0.0, 0.0}, {0.0, 0.0}}
elevator.selection_box = {{0, 0}, {0, 0}}
elevator.collision_mask = {}
elevator.animation = {layers={{
frame_count = 1,
direction_count = 1,
width = 32,
height = 32,
filename = "__Skyscrapers__/graphics/transparent.png",
}}}
elevator.turret_animation = elevator.animation
elevator.guns = {}
table.insert(elevator.flags,"not-on-map")
table.insert(elevator.flags,"not-blueprintable")
elevator.working_sound.sound.volume = 0.0
elevator.working_sound.activate_sound.volume = 0.0
elevator.working_sound.deactivate_sound.volume = 0.0
elevator.light = {{
        type = "oriented",
        minimum_darkness = 0.3,
        picture =
        {
          filename = "__Skyscrapers__/graphics/transparent.png",
          priority = "extra-high",
          flags = { "light" },
          scale = 1,
          width = 32,
          height = 32
        },
        shift = {0, 0},
        size = 1,
        intensity = 1.0,
        color = {r = 1, g = 1, b = 1}
      }}
data:extend({elevator})

data:extend({{
    type = "sound",
    name = "elevator_ding",
    filename = "__base__/sound/programmable-speaker/vibraphone-36.ogg",
    volume = 1.0
}})
local function getPictures(name, sizeX, sizeY, sizeZ, frameCount, animationSpeed)

    res = {layers = {}}
    
    table.insert(res.layers, {
            filename = "__Skyscrapers__/graphics/entity/" ..name.. "/base.png",
            priority = "extra-high",
            width = sizeX*32,
            height = 64,
            shift = {0.0, (-1+sizeY/2) - 0},
            line_length = 1,
            frame_count = frameCount,
            animation_speed = animationSpeed,
            run_mode = 'forward',
            hr_version =
            {
                filename = "__Skyscrapers__/graphics/entity/" ..name.. "/hr-base.png",
                priority = "extra-high",
                width = 2*sizeX*32,
                height = 2*64,
                shift = {0.0, (-1+sizeY/2) - 0},
                line_length = 1,
                frame_count = frameCount,
                animation_speed = animationSpeed,
                run_mode = 'forward',
                picture = hrFileName,
                scale = 0.5,
            }
        })
    for i = 1,sizeZ-1 do
        local tint = { r=1.0, g=1.0, b=1.0, a=0.8 }
        if i*2 < sizeY then tint.a = 1.0 end
                    
        table.insert(res.layers, {
            filename = "__Skyscrapers__/graphics/entity/" ..name.. "/floor.png",
            priority = "extra-high",
            width = sizeX*32,
            height = 64,
            tint = tint,
            shift = {0.0, (-1+sizeY/2) - 2*i},
            line_length = 1,
            frame_count = frameCount,
            animation_speed = animationSpeed,
            run_mode = 'forward',
            hr_version = {
                filename = "__Skyscrapers__/graphics/entity/" ..name.. "/hr-floor.png",
                priority = "extra-high",
                width = 2*sizeX*32,
                height = 2*64,
                tint = tint,
                shift = {0.0, (-1+sizeY/2) - 2*i},
                line_length = 1,
                frame_count = frameCount,
                animation_speed = animationSpeed,
                run_mode = 'forward',
                scale = 0.5,
            }
        })
    end
    table.insert(res.layers, {
            filename = "__Skyscrapers__/graphics/entity/" ..name.. "/top.png",
            priority = "extra-high",
            width = sizeX*32,
            height = 96,
            tint = { r=1.0, g=1.0, b=1.0, a=0.8 },
            shift = {0.0, (-1+sizeY/2) - 2*(sizeZ) -0.5},
            line_length = 1,
            frame_count = frameCount,
            animation_speed = animationSpeed,
            run_mode = 'forward',
            hr_version = {
                filename = "__Skyscrapers__/graphics/entity/" ..name.. "/hr-top.png",
                priority = "extra-high",
                width = 2*sizeX*32,
                height = 2*96,
                tint = { r=1.0, g=1.0, b=1.0, a=0.8 },
                shift = {0.0, (-1+sizeY/2) - 2*(sizeZ) -0.5},
                line_length = 1,
                frame_count = frameCount,
                animation_speed = animationSpeed,
                run_mode = 'forward',
                scale = 0.5,
            }
        })
    return res
    
end

local function makeSingleImageTile(tilename, filename, hrFileName)

    local newMain = 
    {
        {
            picture = filename,
            count = 1,
            size = 1
        }
    }
    if hrFileName then
        newMain = 
        {
            {
                picture = filename,
                count = 1,
                size = 1,
                hr_version =
                {
                    picture = hrFileName,
                    count = 1,
                    scale = 0.5,
                    size = 1
                }
            }
        }
    end

    return {
    name = tilename,
    type = "tile",
    collision_mask =
    {
      "ground-tile",
      "item-layer",
      "object-layer",
    },
    layer = 70,
    variants =
    {
      main = newMain,
      inner_corner =
      {
        picture = "",
        count = 0
      },
      outer_corner =
      {
        picture = "",
        count = 0
      },
      side =
      {
        picture = "",
        count = 0
      }
    },
    map_color={r=0, g=0, b=1},
    ageing=0.0006
  }
end

data:extend({
    makeSingleImageTile("skyscraper-exit", "__base__/graphics/icons/blueprint.png"),
    makeSingleImageTile("skyscraper-stairs-up", "__Skyscrapers__/graphics/tiles/StairsUp.png"),
    makeSingleImageTile("skyscraper-stairs-down", "__Skyscrapers__/graphics/tiles/StairsDown.png"),
    makeSingleImageTile("skyscraper-elevator-entrance-1", "__Skyscrapers__/graphics/tiles/door1/door1.png", "__Skyscrapers__/graphics/tiles/door1/hr-door1.png"),
    makeSingleImageTile("skyscraper-elevator-entrance-2", "__Skyscrapers__/graphics/tiles/door1/door2.png", "__Skyscrapers__/graphics/tiles/door1/hr-door2.png"),
    makeSingleImageTile("skyscraper-elevator-entrance-3", "__Skyscrapers__/graphics/tiles/door1/door3.png", "__Skyscrapers__/graphics/tiles/door1/hr-door3.png"),
    makeSingleImageTile("skyscraper-elevator-entrance-4", "__Skyscrapers__/graphics/tiles/door1/door4.png", "__Skyscrapers__/graphics/tiles/door1/hr-door4.png"),

    {
        type = "item",
        name = "skyscraper-template",
        icon = "__Skyscrapers__/graphics/icons/skyscraper_small.png",
        icon_size = 32,
        flags = {"goes-to-main-inventory"},
        subgroup = "skyscrapers-buildings-mk1",
        order = "a",
        place_result = "skyscraper-template",
        stack_size = 10
    },
    
    {
        type = "recipe",
        name = "skyscraper-template",
        enabled = "true",
        category = "crafting",
        ingredients =
        {
            --{name = "raw-wood", amount = 20},
            --{name = "stone", amount = 20},
        },
        energy_required = 10,
        result = "skyscraper-template",
        result_count = 1
    },
    
    {
        type = "electric-energy-interface",
        name = "skyscraper-template",
        icon = "__Skyscrapers__/graphics/icons/skyscraper_small.png",
        icon_size = 32,
        order="e-a-b",
        flags = {"placeable-neutral", "player-creation", "not-blueprintable", "not-rotatable", "not-deconstructable"},
        minable = {hardness = 0.2, mining_time = 0.5, result = nil},
        max_health = 150,
        corpse = "big-remnants",
        collision_box = {{-4.9, -3.9}, {4.9, 3.9}},
        selection_box = {{-5.0, -4.0}, {5.0, 4.0}},
        energy_source =
        {
            type = "electric",
            buffer_capacity = "10MJ",
            usage_priority = "secondary-input",
            input_flow_limit = "10MW",
            output_flow_limit = "1kW"
        },
        animation = getPictures("skyscraper-template", 10, 8, 5, 1, 0.001),
        charge_cooldown = 30,
        charge_light = {intensity = 0.3, size = 7, color = {r = 1.0, g = 1.0, b = 1.0}},
        discharge_cooldown = 60,
        discharge_light = {intensity = 0.7, size = 7, color = {r = 1.0, g = 1.0, b = 1.0}},
        vehicle_impact_sound =  { filename = "__base__/sound/car-metal-impact.ogg", volume = 0.65 },
        working_sound =
        {
            sound =
            {
                filename = "__base__/sound/accumulator-working.ogg",
                volume = 1
            },
            idle_sound = {
                filename = "__base__/sound/accumulator-idle.ogg",
                volume = 0.4
            },
            max_sounds_per_type = 5
        },
        circuit_wire_connection_point =
        {
            shadow =
            {
                red = {0.984375, 1.10938},
                green = {0.890625, 1.10938}
            },
            wire =
            {
                red = {0.6875, 0.59375},
                green = {0.6875, 0.71875}
            }
        },
        circuit_connector_sprites = circuit_connector_definitions["belt"].sprites,
        circuit_wire_max_distance = 9,
        default_output_signal = {type = "virtual", name = "signal-A"},
    },
    
    {
        type = "container",
        name = "skyscraper-remnants",
        icon = "__base__/graphics/icons/steel-chest.png",
        icon_size = 32,
        order = "nowhere",
        flags = {"placeable-neutral", "player-creation", "not-blueprintable"},
        minable = {mining_time = 1, result = nil},
        max_health = 350,
        corpse = "small-remnants",
        open_sound = { filename = "__base__/sound/metallic-chest-open.ogg", volume=0.65 },
        close_sound = { filename = "__base__/sound/metallic-chest-close.ogg", volume = 0.7 },
        resistances =
        {
            {
                type = "fire",
                percent = 90
            },
            {
                type = "impact",
                percent = 60
            }
        },
        collision_box = {{-1.35, -1.35}, {1.35, 1.35}},
        selection_box = {{-1.5, -1.5}, {1.5, 1.5}},
        fast_replaceable_group = "container",
        inventory_size = 48,
        vehicle_impact_sound =  { filename = "__base__/sound/car-metal-impact.ogg", volume = 0.65 },
        picture =
        {
            filename = "__Skyscrapers__/graphics/entity/skyscraper-remnants/skyscraper-remnants.png",
            priority = "extra-high",
            width = 109,
            height = 102,
            shift = {0, 0}
        },
        circuit_wire_connection_point = circuit_connector_definitions["chest"].points,
        circuit_connector_sprites = circuit_connector_definitions["chest"].sprites,
        circuit_wire_max_distance = default_circuit_wire_max_distance
    },
    {
        type = "technology",
        name = "Skyscrapers_mk1",
        icon = "__Skyscrapers__/graphics/icons/skyscraper_big.png",
        icon_size = 96,
        effects =
        {
            {
                type = "unlock-recipe",
                recipe = "long-pipe-to-ground"
            },
            {
                type = "unlock-recipe",
                recipe = "long-underground-belt"
            },
            {
                type = "unlock-recipe",
                recipe = "long-fast-underground-belt"
            },
        },
        prerequisites = {"SNTD-nixie-tubes-2", "optics", "circuit-network", "logistics-2"},
        unit =
        {
        count = 100,
        ingredients =
        {
        {"science-pack-1", 1},
        {"science-pack-2", 1}
        },
        time = 30
        },
        upgrade = false,
        order = "s-1"
    },
    {
        type = "technology",
        name = "Skyscrapers-elevators",
        icon = "__Skyscrapers__/graphics/tiles/door1/hr-door2.png",
        icon_size = 64,
        effects =
        {
        },
        prerequisites = {"Skyscrapers_mk1", "steel-processing", "electric-engine"},
        unit =
        {
        count = 100,
        ingredients =
        {
        {"science-pack-1", 1},
        {"science-pack-2", 1}
        },
        time = 30
        },
        upgrade = false,
        order = "s-2-1"
    },
    {
        type = "technology",
        name = "Skyscrapers_mk2",
        icon = "__Skyscrapers__/graphics/icons/skyscraper_big.png",
        icon_size = 96,
        effects =
        {
        },
        prerequisites = {"Skyscrapers_mk1", "Skyscrapers-elevators"},
        unit =
        {
        count = 100,
        ingredients =
        {
        {"science-pack-1", 1},
        {"science-pack-2", 1}
        },
        time = 30
        },
        upgrade = false,
        order = "s-2-2"
    },
    {
        type = "technology",
        name = "Skyscrapers-beacons",
        icon = "__Skyscrapers__/graphics/technology/skyscraper-beacon.png",
        icon_size = 128,
        effects =
        {
        },
        prerequisites = {"Skyscrapers_mk2", "effect-transmission"},
        unit =
        {
        count = 200,
        ingredients =
        {
        {"science-pack-1", 1},
        {"science-pack-2", 1},
        {"science-pack-3", 1},
        {"high-tech-science-pack", 1}
        },
        time = 30
        },
        upgrade = false,
        order = "s-3-1"
    },
    {
        type = "technology",
        name = "Skyscrapers_mk3",
        icon = "__Skyscrapers__/graphics/icons/skyscraper_big.png",
        icon_size = 96,
        effects =
        {
        },
        prerequisites = {"Skyscrapers_mk2", "Skyscrapers-beacons"},
        unit =
        {
        count = 500,
        ingredients =
        {
        {"science-pack-1", 1},
        {"science-pack-2", 1},
        {"science-pack-3", 1},
        {"high-tech-science-pack", 1}
        },
        time = 30
        },
        upgrade = false,
        order = "s-3-2"
    },
})

for name,templateData in pairs(skyscrapers) do

    local skyscraperIcons = nil
    if not templateData.unique and templateData.allowedCraftingCategories then 
        skyscraperIcons = {
            {
                icon = "__Skyscrapers__/graphics/icons/skyscraper_small.png",
                scale = 1.0,
                icon_size = 32,
            },
            {
                icon = data.raw.item[templateData.restrictionItemName].icon,
                scale = 0.5,
                shift = {4, -8},
                icon_size = data.raw.item[templateData.restrictionItemName].icon_size,
            }
        }
    end

    local skyscraperEntity = table.deepcopy(data.raw["electric-energy-interface"]["skyscraper-template"])
    skyscraperEntity.name = name
    skyscraperEntity.minable.result = nil --todo figure out how to get a on_player_mined_entity callback when no result is defined. would be better to have everything in remnants
    skyscraperEntity.minable.results = 
    {
        {type="item", name = "skyscraper-floor-"..templateData.tileCraft.."-"..tostring(templateData.X).."-"..tostring(templateData.Y), amount=templateData.Z + 1},
    }
    if templateData.elevator then
        table.insert(skyscraperEntity.minable.results, {type="item", name="elevator-"..tostring(templateData.Z), amount=1})
    end
    
    skyscraperEntity.animation = getPictures(name, templateData.X, templateData.Y, templateData.Z, templateData.numFrames, templateData.animationSpeed)
    if skyscraperIcons then skyscraperEntity.icon = nil skyscraperEntity.icon_size = nil skyscraperEntity.icons = skyscraperIcons end
    skyscraperEntity.collision_box = {{0.1-templateData.X/2.0, 0.1-templateData.Y/2.0}, {-0.1+templateData.X/2.0,-0.1+templateData.Y/2.0}}
    skyscraperEntity.selection_box = {{-templateData.X/2.0, -templateData.Y/2.0}, {templateData.X/2.0, templateData.Y/2.0}}
    skyscraperEntity.energy_source.buffer_capacity = templateData.energyBuffer
    skyscraperEntity.energy_source.input_flow_limit = templateData.energyFlow
    skyscraperEntity.energy_source.output_flow_limit = "0J"
    local restrictionStr = ""
    if templateData.allowedCraftingCategories then for _,v in pairs(templateData.allowedCraftingCategories) do restrictionStr = restrictionStr .. v .. "," end end
    if string.len(restrictionStr) > 0 then restrictionStr = string.sub(restrictionStr, 1, string.len(restrictionStr)-1) end
    if not templateData.unique and not templateData.allowedCraftingCategories then skyscraperEntity.localised_name = {"item-name.skyscraper-unrestricted",tostring(templateData.Z)} end
    if not templateData.unique and templateData.allowedCraftingCategories then skyscraperEntity.localised_name = {"item-name.skyscraper-restricted",tostring(templateData.Z), restrictionStr} end
    if templateData.unique and not templateData.allowedCraftingCategories then skyscraperEntity.localised_name = {"item-name.skyscraper-unique-unrestricted",templateData.uniqueName, tostring(templateData.Z)} end
    if templateData.unique and templateData.allowedCraftingCategories then skyscraperEntity.localised_name = {"item-name.skyscraper-unique-unrestricted",templateData.uniqueName, tostring(templateData.Z), restrictionStr} end
    if not templateData.unique and not templateData.allowedCraftingCategories then skyscraperEntity.localised_description = {"item-description.skyscraper-unrestricted",tostring(templateData.Z)} end
    if not templateData.unique and templateData.allowedCraftingCategories then skyscraperEntity.localised_description = {"item-description.skyscraper-restricted",tostring(templateData.Z), restrictionStr} end
    if templateData.unique and not templateData.allowedCraftingCategories then skyscraperEntity.localised_description = {"item-description.skyscraper-unique-unrestricted",templateData.uniqueDescription, tostring(templateData.Z)} end
    if templateData.unique and templateData.allowedCraftingCategories then skyscraperEntity.localised_description = {"item-description.skyscraper-unique-unrestricted",templateData.uniqueDescription, tostring(templateData.Z), restrictionStr} end
    if templateData.unique then skyscraperEntity.icon = "__Skyscrapers__/graphics/icons/"..name.."/icon.png" end
    if templateData.unique then skyscraperEntity.icon_size = 64 end
    data:extend({skyscraperEntity})

    local skyscraperRecipe = table.deepcopy(data.raw.recipe["skyscraper-template"])
    skyscraperRecipe.name = name
    skyscraperRecipe.result = name
    local numCombinators = templateData.Z
    if templateData.elevator then numCombinators = numCombinators + templateData.Z end
    skyscraperRecipe.ingredients = {
        {name = "skyscraper-floor-"..templateData.tileCraft.."-"..tostring(templateData.X).."-"..tostring(templateData.Y), amount = templateData.Z + 1},
        {name = "small-lamp", amount = templateData.Z},
        {name = "constant-combinator", amount = numCombinators},
        {name = "SNTD-nixie-tube", amount = templateData.Z * 2},
    }
    if templateData.elevator then table.insert(skyscraperRecipe.ingredients, {name = "SNTD-nixie-tube-small", amount = templateData.Z}) end
    if templateData.elevator then table.insert(skyscraperRecipe.ingredients, {name = "elevator-"..tostring(templateData.Z), amount = 1}) end
    if templateData.beacon then table.insert(skyscraperRecipe.ingredients, {name = 
        "skyscraper-beacon-array-"..tostring(templateData.Z).."-"..tostring(templateData.beaconModules).."-"..tostring(templateData.beaconEfficiency),
        amount = 1}) 
    end
    skyscraperRecipe.enabled=false
    data:extend({skyscraperRecipe})
    if not templateData.unique then
    table.insert(data.raw["technology"][templateData.techToUnlock].effects, {
            type = "unlock-recipe",
            recipe = name
        })
    else
        data:extend({{
            type = "technology",
            name = "Skyscrapers_unique_"..name,
            icon = "__Skyscrapers__/graphics/icons/"..name.."/icon.png",
            icon_size = 64,
            effects =
            {
                {
                    type = "unlock-recipe",
                    recipe = name
                }
            },
            prerequisites = {templateData.predecessorTech},
            unit =
            {
            count = templateData.techingredientsCount,
            ingredients = templateData.techingredients,
            time = 30
            },
            upgrade = false,
            order = "s-4",
            localised_name = {"technology-name.skyscraper_unique",templateData.techName},
            localised_description = {"technology-description.skyscraper_unique",templateData.techDescription},
            }})
    end
    local skyscraperItem = table.deepcopy(data.raw.item["skyscraper-template"])
    skyscraperItem.name = name
    skyscraperItem.result = name
    skyscraperItem.place_result = name
    skyscraperItem.localised_name = skyscraperEntity.localised_name
    if not templateData.unique then skyscraperItem.subgroup = templateData.itemCategory end
    if templateData.unique then skyscraperItem.subgroup = "skyscrapers-buildings-unique" end
    if templateData.unique then skyscraperItem.icon = "__Skyscrapers__/graphics/icons/"..name.."/icon.png" end
    if templateData.unique then skyscraperItem.icon_size = 64 end
    if skyscraperIcons then skyscraperItem.icon = nil skyscraperItem.icon_size = nil skyscraperItem.icons = skyscraperIcons end
    data:extend({skyscraperItem})
            
    
    local skyscraperAccumulator = table.deepcopy(data.raw["accumulator"]["accumulator"])
    skyscraperAccumulator.name = name .. "-accumulator"
    skyscraperAccumulator.order = "nowhere"
    skyscraperAccumulator.minable.result = nil
    skyscraperAccumulator.energy_source.buffer_capacity = templateData.energyBuffer
    skyscraperAccumulator.energy_source.input_flow_limit = "0J"
    skyscraperAccumulator.energy_source.output_flow_limit = templateData.energyFlow
    table.insert(skyscraperAccumulator.flags,"not-blueprintable")
    data:extend({skyscraperAccumulator})
    
    if not data.raw.item["skyscraper-floor-"..templateData.tileCraft.."-"..tostring(templateData.X).."-"..tostring(templateData.Y)] then
        data:extend({{
            type = "item",
            name = "skyscraper-floor-"..templateData.tileCraft.."-"..tostring(templateData.X).."-"..tostring(templateData.Y),
            icons = {
                {
                    icon = "__Skyscrapers__/graphics/icons/floor.png",
                    scale = 0.5,
                    icon_size = 64,
                },
                {
                    icon = data.raw.item[templateData.tileCraft].icon,
                    scale = 0.5,
                    shift = {4, -8},
                    icon_size = data.raw.item[templateData.tileCraft].icon_size,
                }
            },
            flags = {"goes-to-main-inventory"},
            subgroup = "skyscrapers-floors",
            order = templateData.tileCraft .. string.format("%08d", templateData.X*templateData.Y),
            localised_name = {"item-name.skyscraper-floor",templateData.tileCraft,tostring(templateData.X),tostring(templateData.Y)},
            localised_description = {"item-name.skyscraper-floor",templateData.tileCraft,tostring(templateData.X),tostring(templateData.Y)},
            stack_size = 100
        }})
        data:extend({{
            type = "recipe",
            name = "skyscraper-floor-"..templateData.tileCraft.."-"..tostring(templateData.X).."-"..tostring(templateData.Y),
            enabled = "false",
            category = "crafting",
            ingredients =
            {
                {name = templateData.tileCraft, amount = templateData.X*templateData.Y},
            },
            energy_required = 5,
            result = "skyscraper-floor-"..templateData.tileCraft.."-"..tostring(templateData.X).."-"..tostring(templateData.Y),
            result_count = 1
        }})
        table.insert(data.raw["technology"]["Skyscrapers_mk1"].effects, {
            type = "unlock-recipe",
            recipe = "skyscraper-floor-"..templateData.tileCraft.."-"..tostring(templateData.X).."-"..tostring(templateData.Y)
        })
    end
    if templateData.elevator and not data.raw.item["elevator-"..tostring(templateData.Z)] then
        data:extend({{
            type = "item",
            name = "elevator-"..tostring(templateData.Z),
            icon = "__Skyscrapers__/graphics/tiles/door1/hr-door2.png",
            icon_size = 64,
            flags = {"goes-to-main-inventory"},
            subgroup = "skyscrapers-elevators",
            order = string.format("%08d", templateData.Z),
            localised_name = {"item-name.elevator",tostring(templateData.Z)},
            localised_description = {"item-name.elevator",tostring(templateData.Z)},
            stack_size = 10
        }})
        local steelCableLength = templateData.Z*4*3 --2:1 elevator ratio would need roughly 4 times the cable length. also 3 meter per floor.
        local steelCableLengthStr = tostring(steelCableLength)
        data:extend({{
            type = "recipe",
            name = "elevator-"..tostring(templateData.Z),
            enabled = "false",
            category = "crafting",
            ingredients =
            {
                {name = "iron-plate", amount = 10},
                {name = "steel-plate", amount = 10},
                {name = "steel-cable-"..steelCableLengthStr, amount = 2},
                {name = "electronic-circuit", amount = templateData.Z + 5},
                {name = "electric-engine-unit", amount = 1},
            },
            energy_required = 8,
            result = "elevator-"..tostring(templateData.Z),
            result_count = 1
        }})
        table.insert(data.raw["technology"]["Skyscrapers-elevators"].effects, {
            type = "unlock-recipe",
            recipe = "elevator-"..tostring(templateData.Z),
        })
        if not data.raw.item["steel-cable-"..steelCableLengthStr] then
            data:extend({{
            type = "item",
            name = "steel-cable-"..steelCableLengthStr,
            icon = "__Skyscrapers__/graphics/icons/steel-wire.png",
            icon_size = 32,
            flags = {"goes-to-main-inventory"},
            subgroup = "skyscrapers-cables",
            order = string.format("%08d", steelCableLengthStr),
            localised_name = {"item-name.steel-cable",steelCableLengthStr},
            localised_description = {"item-name.steel-cable",steelCableLengthStr},
            stack_size = 10
            }})
            data:extend({{
                type = "recipe",
                name = "steel-cable-"..steelCableLengthStr,
                enabled = "false",
                category = "crafting",
                ingredients =
                {
                    {name = "steel-plate", amount = steelCableLength},
                },
                energy_required = 5,
                result = "steel-cable-"..steelCableLengthStr,
                result_count = 1
            }})
            table.insert(data.raw["technology"]["Skyscrapers-elevators"].effects, {
                type = "unlock-recipe",
                recipe = "steel-cable-"..steelCableLengthStr,
            })
        end
    end
    
    if templateData.beacon and not data.raw.item["skyscraper-beacon-array-"..tostring(templateData.Z).."-"..tostring(templateData.beaconModules).."-"..tostring(templateData.beaconEfficiency)] then
        data:extend({{
            type = "item",
            name = "skyscraper-beacon-array-"..tostring(templateData.Z).."-"..tostring(templateData.beaconModules).."-"..tostring(templateData.beaconEfficiency),
            icon = "__base__/graphics/icons/beacon.png",
            icon_size = 32,
            flags = {"goes-to-main-inventory"},
            subgroup = "skyscrapers-beacons",
            order = string.format("%08d-%08d-%08d", templateData.Z,templateData.beaconModules,templateData.beaconEfficiency),
            localised_name = {"item-name.skyscraper-beacon-array",tostring(templateData.Z), tostring(templateData.beaconModules), tostring(templateData.beaconEfficiency)},
            localised_description = {"item-name.skyscraper-beacon-array",tostring(templateData.Z), tostring(templateData.beaconModules), tostring(templateData.beaconEfficiency)},
            stack_size = 10
        }})
        data:extend({{
            type = "recipe",
            name = "skyscraper-beacon-array-"..tostring(templateData.Z).."-"..tostring(templateData.beaconModules).."-"..tostring(templateData.beaconEfficiency),
            enabled = "false",
            category = "crafting",
            ingredients =
            {
                {name = "steel-plate", amount = 30},
                {name = "beacon", amount = templateData.Z},
                {name = "advanced-circuit", amount = 40},
                {name = "processing-unit", amount = 40},
                {name = "copper-cable", amount = templateData.Z * 10},
            },
            energy_required = 30,
            result = "skyscraper-beacon-array-"..tostring(templateData.Z).."-"..tostring(templateData.beaconModules).."-"..tostring(templateData.beaconEfficiency),
            result_count = 1
        }})
        table.insert(data.raw["technology"]["Skyscrapers-beacons"].effects, {
            type = "unlock-recipe",
            recipe = "skyscraper-beacon-array-"..tostring(templateData.Z).."-"..tostring(templateData.beaconModules).."-"..tostring(templateData.beaconEfficiency),
        })
    end

    if templateData.beacon and not data.raw["beacon"]["skyscrapers-beacon-template-"..tostring(templateData.beaconModules).."-"..tostring(templateData.beaconEfficiency)] then
        local concreteBeaconTemplate = table.deepcopy(beaconTemplate)
        concreteBeaconTemplate.name = "skyscrapers-beacon-template-"..tostring(templateData.beaconModules).."-"..tostring(templateData.beaconEfficiency)
        concreteBeaconTemplate.module_specification.module_slots=templateData.beaconModules
        concreteBeaconTemplate.distribution_effectivity=templateData.beaconEfficiency / 10.0
        concreteBeaconTemplate.energy_usage = templateData.beaconEnergy
        data:extend({concreteBeaconTemplate})
        local concreteBeaconInstance = table.deepcopy(beaconInstance)
        concreteBeaconInstance.name = "skyscrapers-beacon-instance-"..tostring(templateData.beaconModules).."-"..tostring(templateData.beaconEfficiency)
        concreteBeaconInstance.module_specification.module_slots=templateData.beaconModules
        concreteBeaconInstance.distribution_effectivity=templateData.beaconEfficiency / 10.0
        data:extend({concreteBeaconInstance})
    end
    
end
data.raw["electric-energy-interface"]["skyscraper-template"] = nil
data.raw.recipe["skyscraper-template"] = nil
data.raw.item["skyscraper-template"] = nil

for name,details in pairs(premadeSkyscrapers) do

    local baseTemplate = skyscrapers[details.base]
    local ingredientMap = {}
    for i = 1, baseTemplate.Z do
        local floorEntities = nil
        if i == 1 then 
            floorEntities = details.floor1 
        elseif i == baseTemplate.Z then
            floorEntities = details.floorTop
        else
            floorEntities = details.floorN
        end
        for _,v in pairs(floorEntities) do
            if not ingredientMap[v.name] then ingredientMap[v.name] = 0 end
            ingredientMap[v.name] = ingredientMap[v.name] + 1
        end
    end
    local ingredients = {}
    for k,v in pairs(ingredientMap) do
        table.insert(ingredients, {name = k, amount = v})
    end
    table.insert(ingredients, {name = details.base, amount = 1})

    data:extend({
        {
            type = "item",
            name = name,
            localised_name = {"item-name.skyscraper-prebuild",details.textName},
            localised_description = {"item-name.skyscraper-prebuild",details.textDescription},
            icons = {
                {
                    icon = "__Skyscrapers__/graphics/icons/skyscraper_small.png",
                    scale = 1.0,
                    icon_size = 32,
                },
                {
                    icon = data.raw.item[details.itemName].icon,
                    scale = 0.5,
                    shift = {4, -8},
                    icon_size = data.raw.item[details.itemName].icon_size,
                }
            },
            flags = {"goes-to-quickbar"},
            subgroup = "skyscrapers-buildings-premade",
            order = "c[splitter]-a[splitter]",
            place_result = details.base,
            stack_size = 50
        },
        {
            type = "recipe",
            name = name,
            enabled = "false",
            category = "crafting",
            ingredients = ingredients,
            energy_required = 20,
            result = name,
            result_count = 1
        },
    })
    
    table.insert(data.raw["technology"][details.techToUnlock].effects, {
        type = "unlock-recipe",
        recipe = name
    })
    
end